(************************************************************* 
			Mini  Java 
 *************************************************************
 Auteur :	Guy Cousineau
 Fichier :	javasyntax.ml
 *************************************************************
 Description :	syntaxe abstraite de Mini Java en Caml
 *************************************************************)

type  'a option = Nopt | Some of 'a;;
type   status = InstStatus | ClassStatus;;


(* ***********************************************************
 * Programme Java 
 * *********************************************************** *) 

type javaprog = JavaProg of javaclassdecl list


(* ***********************************************************
 * Declarations de classe
 * *********************************************************** *) 

and javaclassdecl = {classname: string; 
                     classfather: string option;
                     fielddecls: javafielddecl list;
                     methdecls: javamethdecl list}


(* ***********************************************************
 * Declarations et definitions de methode
 * *********************************************************** *) 

and javamethdecl = {methname: string; 
                    methrestype: javatype;
                    methparams: (string * javatype) list;
                    methstatus: status;
                    methbody: javastat}


(* ***********************************************************
 * Declarations de champs et variables
 * *********************************************************** *) 

and javafielddecl = {fieldname: string;
                     fieldtype: javatype;
                     fieldstatus: status;
                     fieldinit: javaexp option}

and javavardecl = {varname: string;
                   vartype: javatype;
                   varinit: javaexp option}

(* ***********************************************************
 * Declarations de base
 * *********************************************************** *) 

and javatype = JavaVoidType 
             | JavaBoolType | JavaIntType
             | JavaIdType of javaident
             | JavaNullType

and javaident = string


(* ***********************************************************
 * Expressions
 * *********************************************************** *) 

and javaexp =  JavaConstExp of javaconst
            | JavaThisExp
            | JavaVarExp of javaident
            | JavaNewClassExp of javaident
            | JavaMethCall of javaexp * string * (javaexp list)
            | JavaCall of string *  (javaexp list)
            | JavaFieldExp of javaexp * string
            | JavaUnopExp of javaunop * javaexp
            | JavaBinopExp of javabinop * javaexp * javaexp
            | JavaAssign of javaexp * javaexp

and javaconst = JavaInt of int 
              | JavaTrue
              | JavaFalse
              | JavaNull

and javaunop = JavaUMOINS | JavaNOT
and javabinop = JavaAND | JavaOR | JavaPLUS | JavaMOINS | JavaMULT | JavaDIV 
              | JavaEGAL | JavaNE | JavaLT | JavaGT | JavaLE | JavaGE

(* ***********************************************************
 * Blocks et instructions
 * *********************************************************** *)

and javastat = JavaBlock of (javavardecl list) * (javastat list)
          | JavaComput of javaexp
          | JavaIf of javaexp * javastat * javastat
          | JavaFor of javaexp * javaexp * javaexp * javastat
          | JavaReturn of javaexp option
          | JavaNop

;;

(* ***********************************************************
 * DEBUG
 * *********************************************************** *)

let noline = ref 1;;







