(*************************************************************
                        Mini Java
 *************************************************************
 Auteur :       **** A COMPLETER ****
 Fichier :      javatyper.ml
 *************************************************************
 Description :  typage de Mini Java en OCaml selon les regles 
                indiquees dans regles_typage.ps
 *************************************************************)


(*************************************************************
  Ouverture des modules utilises
 *************************************************************)
open Analyseur;;
open Javasyntax;;

(*************************************************************
  Declaration de l'exceptions levee en cas d'erreur de typage
 *************************************************************)
exception Erreur_typage of string;;


(*************************************************************
  Fonctions pour la definition de la relation de sous-typage
 *************************************************************)
(**
  Renvoie true si classname1 est une sousclasse de classname2.
   @param table resultat de @see table_ancetres ((string * string list) list)
   @param classname1 nom de la sous-classe (string)
   @param classname1 nom de la super-classe (string)
   @return      un boolean
 *)
let rec sousclasse table classname1 classname2 =
  List.mem classname2 (List.assoc classname1 table)
;;

(**
  Renvoie true si ty1 est un soustype de ty2. On considere que
  les inclusions proviennent de la hierarchie de classes et pas
  des types de base.
   @param table resultat de @see table_ancetres ((string * string list) list)
   @param ty1   nom du sous-type (javatype)
   @param ty2   nom du super-type (javatype)
   @return      un boolean
 *)
let rec soustype table  ty1 ty2 =
  (ty1=ty2) ||
  match (ty1,ty2)
  with     
    JavaIdType classname1 , JavaIdType "Object"
    -> true (* toute classe est sousclasse d'Object *)
  | JavaIdType classname1 , JavaIdType classname2
    -> sousclasse table classname1 classname2
  | JavaNullType , JavaIdType _
    -> true
  | _,_   -> false
;;

(*************************************************************
  Fonctions pour la construction des tables d'ancetres
  *************************************************************
  Pour les classes
 *************************************************************)
(**
  Calcule la definition de la classe Object
  *)
let object_class = cdecl_of_string "class Object{}";;

(* Declaration de classe bidon pour getclassdef :  effacer *)
let classdecl_bidon = 
  { classname = ""; 
   classfather = Nopt; 
   fielddecls = []; 
   methdecls = [] }
;;

(**
  Renvoie la definition d'une classe. Leve une exception si la classe 
  n'existe pas.
   @param cname nom de la classe (string)
   @param prog  programme a typer (javaprog)
   @return      (methdecl)
  *)
let getclassdef cname (JavaProg(classlist)) = 
 (**** A COMPLETER ****) classdecl_bidon
;;

(**
  Calcule la liste des ancetres pour la classe dont le nom est 
  donn en parametre. Leve exception si heritage recursif.
   @param cname nom de la classe (string)
   @param prog  programme a typer (javaprog)
   @return      liste des ancetres (string list)
 *)
let  ancetres cname prog = (**** A COMPLETER ****) []
;;

(**
  Calcule la table d'ancetres pour toutes les classes du 
  programme.
   @param prog  programme a typer (javaprog)
   @return      table des ancetres, pour chaque classe
                sa liste d'ancetres ((string * string list) list)
 *)
let table_ancetres (JavaProg(classlist) as prog) =
  List.map 
    (function {classname=c} -> c, ancetres c prog)
    classlist
;;

let father cname table = List.hd (List.assoc cname table);;


(*************************************************************
  Fonctions pour la rsolution de la surcharge a l'appel de methode
 *************************************************************)
(**
  Renvoie true si la liste des arguments a l'appel est
  compatible (par sous-typage) avec la liste des parametres, 
  c'est-a-dire :
  - la liste des arguments a l'appel a la mm longueur que celle des 
    parametres formels
  - chaque type argument est le sous-type de son correspondant dans la
    liste des parametres
   @param table      resultat de @see table_ancetres ((string * string list) list)
   @param argtypes   liste des types des arguments (javatype list)
   @param paramtypes liste des types des parametres formels  (javatype list)
   @return           boolean
 *)
let signature_acceptable table argtypes paramtypes =
  try
    List.for_all (function (x,y) -> soustype table  x y)
                 (List.combine argtypes paramtypes)
  with 
    Invalid_argument "List.combine"
    -> false (* car l'exception est levee si les deux listes n'ont pas la mm longueur *)
;;

(**
  Renvoie la liste des types des parametres formels d'une methode a
  partir de la definition de la methode.
   @param methdef la definition de la methode (javamethdecl)
   @return        liste de types des parametres formels (javatype list)
 *)
let signature methdef =  (**** A COMPLETER ****) []

;;

(**
  Cacule la liste des methodes qui sont definies dans la hierarchie d'une classe
  et qui ont le mm nom et les parametres compatibles avec les arguments.
   @param prog       programme type (javaprog)
   @param table      resultat de @see table_ancetres ((string * string list) list)
   @param mname      nom de la methode cherchee (string)
   @param argtypes   liste des types des arguments d'appel (javatype list)
   @param cname      nom de la classe d'appel (string)
   @return           liste (classe de definition, methode)
                     (string * javamethdecl) list
 *)
let methodes_compatibles prog table mname argtypes cname =
  (* on construit la hierarchie de classes dans lesquelles on cherche la methode *)
  let hierarchie = cname::List.assoc cname table in
  (* on definit une fonction locale renvoie les methodes compatibles
     dans la classe c *)
  let methodes_compatibles_dans_classe c =
    let methlist =
          List.filter 
             (function {methname=mn} as methdef
               -> mn=mname &&  signature_acceptable table argtypes (signature methdef))
             ((getclassdef c prog).methdecls)
     in List.map (function meth -> c, meth) methlist
   in
    List.flatten 
    (List.map methodes_compatibles_dans_classe  hierarchie)
;;


(**
  Calcule la definition d'une methode, la plus specifique par rapport
  a la classe demandee a l'appel et aux arguments. Leve une exception
  si la definition n'as pas ete trouvee.
   @param prog       programme type (javaprog)
   @param table      resultat de @see table_ancetres ((string * string list) list)
   @param classname  nom de la classe d'appel (string)
   @param methname   nom de la methode appelee (string)
   @param argtypes   liste des types des arguments d'appel (javatype list)
   @return           la definition de la methode (javamethdecl)
 *)
let trouver_methode prog table classname methname argtypes =
  (* on construit la liste des methodes compatibles dans la hierarchie de classname *)
  let methodes = methodes_compatibles prog table methname argtypes classname in
  try
    (* on cherche dans methodes celle qui est la plus spcifique, cad,
       celle dont la liste classe de dfinition::signature est un sous-type
       (@see signature_acceptable) par toute les autres methodes.
     *)
    snd (List.find
      (function (clname,methdef) ->
        List.for_all
          (function (clname1,methdef1) -> 
	    signature_acceptable table
              (JavaIdType clname::signature methdef)
              (JavaIdType clname1::signature methdef1))
          methodes
      )
     methodes)
 with
    Not_found -> raise(Erreur_typage("La methode " ^ methname ^ 
                           " n'existe pas ou est definie de facon ambigue" ^
                           "dans la classe " ^ classname))
;;

(*************************************************************
  Fonctions utiles pour detecter les doublons
 *************************************************************)
 (* Verifie qu'une liste ne contient pas de doublon *)
 (* Sert a verifier l'unicite des champs dans une classe *)
 (* et des parametres dans une methode *)
 let is_set l =
    List.for_all 
	  (function id -> 
	    (List.length(List.filter (function id1 -> id=id1) l) = 1)
	  )
	  l;;
	  
(* Liste des methodes definies dans la hierarchie de la classe cname *)
(* ayant un nom et une signature donnees *)
let methodes_de_meme_signature prog table mname signature cname =
  (* on construit l'hierarchie de classes dans laquel on cherche la methode *)
  let hierarchie = cname::List.assoc cname table in
  (* on definit une fonction locale renvoie les methodes de signature "signature"
     dans la classe cname *)
  let methodes_de_meme_signature_dans_classe mname argtypes cname =
    List.filter (function md -> md.methname = mname
                             && List.map snd md.methparams = signature) 
                (getclassdef cname prog).methdecls in
    List.flatten 
    (List.map (methodes_de_meme_signature_dans_classe mname signature) hierarchie)
;;

(* Sert a verifier que lorsqu'on redefinit une methode avec la meme signature *)
(* le type resultat n'est pas change *)
let resultat_unique prog table mname signature cname =
 let methlist = methodes_de_meme_signature prog table mname signature cname in
 let reslist = List.map (function md -> md.methrestype) methlist in
   match reslist
   with ([]|[_]) -> true
     |  a::l -> List.for_all (function x -> x=a) l;;
     
	  
(*************************************************************
  Typage d'une expression
 *************************************************************)
 let typer_unopexp op ty =
     match (op,ty)
     with (JavaUMOINS,JavaIntType) ->  (**** A COMPLETER ****) JavaVoidType
      | (JavaNOT,JavaBoolType)     ->  (**** A COMPLETER ****) JavaVoidType
      |  _ -> raise (Erreur_typage("Mauvais argument pour operateur unaire"))  
      ;;
 
 
  let typer_binopexp op ty1 ty2 =
     match (op,ty1,ty2)
     with JavaAND,JavaBoolType,JavaBoolType -> (**** A COMPLETER ****) JavaVoidType
      | JavaOR,JavaBoolType,JavaBoolType    -> (**** A COMPLETER ****) JavaVoidType
      | JavaPLUS,JavaIntType,JavaIntType    -> (**** A COMPLETER ****) JavaVoidType
      | JavaMOINS,JavaIntType,JavaIntType   -> (**** A COMPLETER ****) JavaVoidType
      | JavaMULT,JavaIntType,JavaIntType    -> (**** A COMPLETER ****) JavaVoidType
      | JavaDIV,JavaIntType,JavaIntType     -> (**** A COMPLETER ****) JavaVoidType
      | JavaEGAL,JavaIntType,JavaIntType    -> (**** A COMPLETER ****) JavaVoidType
      | JavaEGAL,JavaBoolType,JavaBoolType  -> (**** A COMPLETER ****) JavaVoidType
      | JavaNE,JavaIntType,JavaIntType      -> (**** A COMPLETER ****) JavaVoidType
      | JavaNE,JavaBoolType,JavaBoolType    -> (**** A COMPLETER ****) JavaVoidType
      | JavaLT,JavaIntType,JavaIntType      -> (**** A COMPLETER ****) JavaVoidType
      | JavaGT,JavaIntType,JavaIntType      -> (**** A COMPLETER ****) JavaVoidType
      | JavaLE,JavaIntType,JavaIntType      -> (**** A COMPLETER ****) JavaVoidType
      | JavaGE,JavaIntType,JavaIntType      -> (**** A COMPLETER ****) JavaVoidType
      |  _ -> raise (Erreur_typage("Mauvais argument pour operateur binaire")) 
      ;;

(* Est-ce que l'expression  gauche peut tre affecte ? *)     

let is_assignable e =
  match e
  with (JavaVarExp _ | JavaFieldExp _) -> true
   |     _  -> false;;

(* Renvoie la dfinition du champ et le nom de la classe dans laquelle 
   apparait le nom du champ *)
   
let get_field_def prog table cname fname =
   let hierarchie = cname::List.assoc cname table in
     let rec get fname l =
     match l with []  ->  raise Not_found
       |  c::l' -> (try let fd = List.find (function fd -> fd.fieldname=fname) 
                                            (getclassdef c prog).fielddecls
                        in fd,c
                    with Not_found -> get fname l')
     in get fname hierarchie;;
     

     
(*
  Typage d'une expression : cf les regles de typage.
   @param prog   programme type (javaprog)
   @param table  table d'heritage ((string * string list) list)
   @param cname  classe de definition (string)
   @param ctx    contexte de typage pour les variables ((string * javatype) list)
   @param status statut de la methode dans laquelle apparait l'expr
   @param exp    expression (javaexp)
   @return       type de l'expression (javatype)
 *)
let rec typer_exp prog table cname ctx status exp = 
  match exp with
    JavaConstExp (JavaInt _)            ->    (**** A COMPLETER ****)  JavaVoidType
  | JavaConstExp (JavaTrue | JavaFalse) ->    (**** A COMPLETER ****)  JavaVoidType
  | JavaConstExp JavaNull ->                  (**** A COMPLETER ****)  JavaVoidType
  | JavaThisExp ->                            (**** A COMPLETER ****)  JavaVoidType 
  | JavaVarExp v ->                           (**** A COMPLETER ****)  JavaVoidType
  | JavaNewClassExp id ->                     (**** A COMPLETER ****)  JavaVoidType 
  | JavaFieldExp(e, fname) ->                 (**** A COMPLETER ****)  JavaVoidType
  | JavaMethCall(e, mname, explist) ->        (**** A COMPLETER ****)  JavaVoidType
  | JavaCall(mname, explist) ->               (**** A COMPLETER ****)  JavaVoidType

  | JavaUnopExp(op, e)       ->   
      typer_unopexp op (typer_exp prog table cname ctx status e)

  | JavaBinopExp(op, e1, e2) ->               
      typer_binopexp   op  (typer_exp prog table cname ctx status e1)
                          (typer_exp prog table cname ctx status e2)

  | JavaAssign(e1, e2) ->                     (**** A COMPLETER ****)  JavaVoidType
;;

(*************************************************************
  Typage d'une declaration de variable
 *************************************************************)
(**
  Typage d'une declaration de variables :
   - le type de declaration et celui de l'expression (quand il y a)
   doivent correspondre.
   @param prog   programme type (javaprog)
   @param table  table d'heritage ((string * string list) list)
   @param name   classe de definition (string)
   @param ctx    contexte de typage pour les variables ((string * javatype) list)
   @param status statut de la methode dans laquelle apparait l'expr
   @param vardecl declaration a typer (javavardecl)
   @return       le nouvel contexte ((string * javatype) list)
 *)
let typer_vardecl prog table cname ctx status vardecl =
  (* on calcule le type de l'expression d'initialisation *)

         (**** A COMPLETER ****)

  (* on verifie que ce type est un sous-type du type declare *)

         (**** A COMPLETER ****)

  (* on verifie que la variable n'est pas deja dans le contexte *)

         (**** A COMPLETER ****) []
;;
(* on applique recoursivement la fonction ci-dessus *)
let rec typer_vardecllist prog table cname ctx status varlist =
  match varlist with 
    [] -> ctx
  | vardecl::l ->
      let newctx = typer_vardecl prog table cname ctx status vardecl in
      typer_vardecllist prog table cname newctx status l
;;
 

(*************************************************************
  Typage d'une instruction
 *************************************************************)

(**
  Typage d'une instruction : cf regles de typage
   @param prog   programme type (javaprog)
   @param table  table d'heritage ((string * string list) list)
   @param cname  classe de definition (string)
   @param ctx    contexte de typage pour les variables ((string * javatype) list)
   @param status statut de la methode dans laquelle apparait l'instruction
   @param stat   instruction a typer (javastat)
   @return       rien (unit)
 *)
let rec typer_stat prog table cname ctx status stat =
  match stat with
    JavaBlock (varlist, statlist) ->          (**** A COMPLETER ****) ()
  | JavaComput e ->                           (**** A COMPLETER ****) ()
  | JavaIf(e, sthen, selse) ->                (**** A COMPLETER ****) ()
  | JavaFor(einit, econd, estep, s) ->        (**** A COMPLETER ****) ()
  | JavaReturn(Some e) ->                     (**** A COMPLETER ****) ()
  | JavaReturn(Nopt) ->                       (**** A COMPLETER ****) ()
  | JavaNop ->                                (**** A COMPLETER OU PAS? ****) ()
;;

(*************************************************************
  Typage d'une declaration de champ
 *************************************************************)
(**
  Typage d'une declaration de champ :
   - le type de declaration et celui de l'expression (quand il y a)
   doivent correspondre.
   @param prog   programme type (javaprog)
   @param table  table d'heritage ((string * string list) list)
   @param name   classe de definition (string)
   @param fielddecl declaration a typer (javafielddecl)
   @return       rien (unit)
 *)
let typer_fielddecl prog table name fielddecl =
  (* on calcule le type de l'expression d'initialisation *)

        (**** A COMPLETER ****)

  (* on verifie que ce type est sous-type de celui declare *)

        (**** A COMPLETER ****)
    ()
;;


(*************************************************************
  Typage d'une definition de methode
 *************************************************************)
(**
  Typage d'une declaration de methode : 
   - les noms des parametres doivent etre differents
   - le type du resultat et celui de l'instruction (quand il y a)
   doivent correspondre.
   @param prog   programme type (javaprog)
   @param table  table d'heritage ((string * string list) list)
   @param cname  classe  de definition (string)
   @param methdecl declaration a typer (javamethdecl)
   @return       rien (unit)
 *)
let typer_methdecl prog table cname methdecl =
  (* on verifie que les noms des parametres sont differents *)

        (**** A COMPLETER ****)

   (* On regarde si des methodes de meme nom et meme signature existent deja *)

        (**** A COMPLETER ****)

  (* on type le corps dans le contexte forme par les parametres  et le type de return *)

	() (**** A COMPLETER ****)
;;



(*************************************************************
  Typage des declarations de classe
 *************************************************************)

(**
  Typage d'une declaration de classe : 
   @param prog   programme type (javaprog)
   @param table  table d'heritage ((string * string list) list)
   @param cdecl  declaration a typer (javainterdecl)
   @return       bool
 *)
let typer_classdef (JavaProg(classlist) as prog) table cdecl =
  (* le champ classfather a ete deja verifie par @see table_ancetres *)
  (* on verifie les definitions de champs *)
  (* On verifie d'abord qu'un champ n'est pas defini deux fois *)
  let fnames = List.map (function fd -> fd.fieldname) cdecl.fielddecls
  in if not (is_set fnames)
   then raise(Erreur_typage("Un champ est defini plusieur fois dans la classe " ^ cdecl.classname))
   else
  let fields = 
    (List.map 
       (function fielddecl -> 
	 typer_fielddecl prog table cdecl.classname fielddecl)
       cdecl.fielddecls
    ) 
  in 
  (* on verifie les definitions de methodes *)
  let methods = 
    (List.map
       (function methdecl -> 
	 typer_methdecl prog table cdecl.classname methdecl)
       cdecl.methdecls
    ) 
  in
  true
;;


(*************************************************************
  Typage d'un programme
 *************************************************************)
(**
  Typage d'un programme par typage des declarations d'interfaces 
  et de classes. Les fonctions qu'elle appelle peuvent lever
  une exception Erreur_typage.
   @param prog       programme type (javaprog)
   @return           bool
 *)
let typer_prog (JavaProg(classlist) as prog) =
  (* on construit la table de la hierarchie des classes *)
  (* cette construction verifie en mm temps que l'heritage est correct,
     cad, que chaque nom de classe  est defini et
     que l'heritage n'est pas recursif *)
  let  table_c = table_ancetres prog in
   List.for_all
     (function classdef -> typer_classdef prog table_c classdef)
     classlist
;;
